# -*- encoding:utf8 -*-

tagset = {
  "Noun": "명사",
  "Pronoun": "대명사",
  "Numeral": "수사",
  "Verb": "동사",
  "Adjective": "형용사",
  "Determiner": "관형사",
  "Adverb": "부사",
  "Exclamation": "감탄사",
  "Josa": "조사",
  "Symbol": "기호"
}